using System;
using System.Collections.Generic;
using System.Text;
using BEPUphysics;
using Microsoft.Xna.Framework;
using System.Diagnostics;
using Microsoft.Xna.Framework.Input;

namespace BEPUphysicsDemos
{
    /// <remarks>
    /// Handles a character's movement and interaction with its surroundings.
    /// </remarks>
    public class CharacterController:Updateable
    {
        /// <summary>
        /// Shape of the character's body.
        /// </summary>
        public Capsule bodyShape;
        /// <summary>
        /// Exists in the space to collect nearby entities for use in testing new step-up positions for validity.
        /// </summary>
        Entity bodyDetector;
        /// <summary>
        /// Exists in the space to collect support candidates and to ensure that other parts of the engine 'know' about the character's feet;
        /// for example, the detector shape would trigger a StaticTriangleGroup to create triangles for support before the character itself hit the triangles, allowing for more robust movement.
        /// </summary>
        Entity feetDetector;
        /// <summary>
        /// Offset from the center position of the body shape to the center of the feet detector shape.
        /// </summary>
        Vector3 feetDetectorOffset;
        /// <summary>
        /// Maximum height up or down which the character can step.
        /// </summary>
        float maximumStepHeight;
        /// <summary>
        /// Offset from the center position of the bodyShape to the starting position of the feet ray.
        /// </summary>
        Vector3 feetOffset;
        /// <summary>
        /// Shape used for identifying supports when the raycast fails to do so.
        /// </summary>
        Cylinder feetShape;

        /// <summary>
        /// List of entities near the body, used in determining if a step-up location is valid.
        /// </summary>
        List<Entity> nearbyEntities = new List<Entity>();
        /// <summary>
        /// List of entities near the feet that could be supporting the object.
        /// </summary>
        List<Entity> supportCandidates = new List<Entity>();
        /// <summary>
        /// Entity on top of which the character is standing.
        /// </summary>
        Entity supportEntity;
        /// <summary>
        /// Location of the impact between the swept feet shape and the support entity.
        /// </summary>
        Vector3 supportLocation;
        /// <summary>
        /// Normal at the supporting surface.
        /// </summary>
        Vector3 supportNormal;

        /// <summary>
        /// Maximum angle off the horizontal a surface can be before it can no longer provide support for the character.
        /// While on a steep surface, the character will act as if airborne and slide.
        /// </summary>
        float maximumSlope;
        /// <summary>
        /// Maximum speed the character can attain under its own power.
        /// </summary>
        float maximumSpeed;
        /// <summary>
        /// Rate of increase in velocity per second when trying to move.
        /// </summary>
        float acceleration;
        /// <summary>
        /// Rate of decrease in velocity per second when not trying to move.
        /// </summary>
        float deceleration;
        /// <summary>
        /// Speed at which the character leaves the ground when it jumps.
        /// </summary>
        float jumpSpeed;

        /// <summary>
        /// Rate at which the character can change its velocity while airborne.
        /// </summary>
        float airAcceleration;
        /// <summary>
        /// Maximum speed that the character can achieve under its own power while airborne.
        /// </summary>
        float maximumAirSpeed;

        /// <summary>
        /// Maximum speed that the character can counteract by pulling itself towards the ground.
        /// </summary>
        float maximumGlueSpeed;
        /// <summary>
        /// Fraction of force used to move the character to apply to the support entity.
        /// </summary>
        float supportEntityForceApplicationFactor;
        /// <summary>
        /// Maximum slope of the support surface past which the entity will begin experiencing slowdown due to gravity.
        /// </summary>
        float penaltySlope;
        /// <summary>
        /// Maximum slope that the character can step to or from.
        /// </summary>
        float maximumStepSlope;
        /// <summary>
        /// Whether or not the character is flying through the air.
        /// </summary>
        bool isAirborne;

        /// <summary>
        /// Normalized direction in which the character should try to move.
        /// </summary>
        public Vector2 movementDirection = Vector2.Zero;

        /// <summary>
        /// Gets and sets the position of the character.
        /// </summary>
        public Vector3 position
        {
            get
            {
                return bodyShape.centerPosition;
            }
            set
            {
                moveTo(value);
            }
        }

        /// <summary>
        /// Creates the character controller.
        /// </summary>
        /// <param name="position">Position to center the character controller's body.</param>
        /// <param name="bodyHeight">Height of the character.</param>
        /// <param name="bodyRadius">Radius of the character.</param>
        /// <param name="mass">Mass of the character.</param>
        /// <param name="skinWidth">Width of the skin (collision margin) on the character's body.</param>
        /// <param name="maximumStepHeight">Maximum height up or down which the character can step.</param>
        /// <param name="acceleration">Rate of increase in velocity per second when trying to move.</param>
        /// <param name="deceleration">Rate of decrease in velocity per second when not trying to move.</param>
        /// <param name="maximumSpeed">Maximum speed the character can attain under its own power.</param>
        /// <param name="maximumSlope">Maximum angle off the horizontal a surface can be before it can no longer provide support for the character.
        /// While on a steep surface, the character will act as if airborne and slide.</param>
        /// <param name="jumpSpeed">Speed at which the character leaves the ground when it jumps.</param>
        /// <param name="airAcceleration">Rate at which the character can change its velocity while airborne.</param>
        /// <param name="maximumAirSpeed">Maximum speed that the character can achieve under its own power while airborne.</param>
        /// <param name="maximumGlueSpeed">Maximum speed that the character can counteract by pulling itself towards the ground.</param>
        /// <param name="supportEntityForceApplicationFactor">Fraction of force used to move the character to apply to the support entity.</param>
        /// <param name="penaltySlope">Maximum slope of the support surface past which the entity will begin experiencing slowdown due to gravity.</param>
        /// <param name="maximumStepSlope">Maximum slope that the character can step to or from.</param>
        public CharacterController(Vector3 position, float bodyHeight, float bodyRadius, float mass, float skinWidth, float maximumStepHeight, float acceleration, float deceleration, float maximumSpeed, float maximumSlope, float jumpSpeed, float airAcceleration, float maximumAirSpeed, float maximumGlueSpeed, float supportEntityForceApplicationFactor, float penaltySlope, float maximumStepSlope)
        {
            //note that the length of the capsule can become negative if the step height and radius are too high.
            //Changing the body to instead use a cylinder would bypass this issue, 
            //though then stuff can fall on your head and not roll off- producing a silly result.  Manually pushing it off would work fine.
            bodyShape = new Capsule(position, bodyHeight - maximumStepHeight - bodyRadius * 2, bodyRadius, mass);
            bodyShape.isAlwaysActive = true;
            bodyShape.localInertiaTensorInverse = Toolbox.zeroMatrix; //Prevents all rotation of the body.
            bodyShape.friction = 0;
            bodyShape.collisionMargin = skinWidth;
            this.maximumStepHeight = maximumStepHeight;
            Debug.WriteLine(bodyShape.length);
            Debug.WriteLine(bodyShape.radius);
            feetOffset = new Vector3(0, -bodyShape.length / 2 - bodyRadius, 0);
            feetShape = new Cylinder(Vector3.Zero, 0, bodyShape.radius);
            feetDetectorOffset = new Vector3(0, -bodyShape.length / 2 - bodyRadius - maximumStepHeight, 0);
            feetDetector = new Box(position + feetDetectorOffset, bodyRadius * 2, maximumStepHeight * 2, bodyRadius * 2, 100000);
            feetDetector.isTangible = false; //This only needs to detect other objects, not collide with them.
            feetDetector.nonCollidableEntities.Add(bodyShape); //Avoid creating a controller with the body shape for no reason.
            //The detector is physical for the purpose of creating controllers with nonphysical objects (two nonphysical objects will not create a controller)
            //so to keep it oriented correctly at all times, prevent any rotation.
            feetDetector.localInertiaTensorInverse = Toolbox.zeroMatrix;

            bodyDetector = new Box(bodyShape.centerPosition, (bodyRadius + skinWidth) * 2, bodyShape.length + bodyRadius * 2 + maximumStepHeight * 2, (bodyRadius + skinWidth) * 2, 100000);
            bodyDetector.isTangible = false;
            bodyDetector.nonCollidableEntities.Add(bodyShape);
            bodyDetector.nonCollidableEntities.Add(feetDetector);
            bodyDetector.localInertiaTensorInverse = Toolbox.zeroMatrix;

            this.acceleration = acceleration;
            this.deceleration = deceleration;
            this.maximumSpeed = maximumSpeed;
            this.maximumSlope = maximumSlope;
            this.jumpSpeed = jumpSpeed;
            this.airAcceleration = airAcceleration;
            this.maximumAirSpeed = maximumAirSpeed;
            this.maximumGlueSpeed = maximumGlueSpeed;
            this.supportEntityForceApplicationFactor = supportEntityForceApplicationFactor;
            this.penaltySlope = penaltySlope;
            this.maximumStepSlope = maximumStepSlope;
            isUpdating = false;
        }

        /// <summary>
        /// Updates the character's state.
        /// </summary>
        /// <param name="dt">Time since last frame in simulation seconds.</param>
        /// <param name="timeScale">Factor of the speed of the simulation.</param>
        /// <param name="timeSinceLastFrame">Time since last frame in real seconds.</param>
        public override void updateAtEndOfUpdate(float dt, float timeScale, float timeSinceLastFrame)
        {
            supportCandidates.Clear();
            //Assemble the list of candidates from the feet detector's controller list.
            foreach (Controller c in feetDetector.controllers)
            {
                if (c.colliderA == feetDetector)
                {
                    if (c.colliderB.isTangible)
                        supportCandidates.Add(c.colliderB);
                }
                else
                    if (c.colliderA.isTangible)
                        supportCandidates.Add(c.colliderA);
            }
            nearbyEntities.Clear();
            //Assemble the list of nearby entities from the body detector's controller list.
            foreach (Controller c in bodyDetector.controllers)
            {
                if (c.colliderA == bodyDetector)
                {
                    if (c.colliderB.isTangible)
                        nearbyEntities.Add(c.colliderB);
                }
                else
                    if (c.colliderA.isTangible)
                        nearbyEntities.Add(c.colliderA);
            }
            findSupport();
            if (!isAirborne)
            {
                //It's on the ground.


                //Put the movement direction into the surface's space.
                Vector3 movementDirection3d = new Vector3(movementDirection.X, 0, movementDirection.Y);
                Vector3 axis = Vector3.Cross(Vector3.Up, supportNormal);
                float axisLength = axis.Length();
                if (axisLength > 0)
                {
                    float supportSlope = (float)Math.Min(penaltySlope, Math.Asin(Math.Min(1, axisLength)));
                    movementDirection3d = Vector3.Transform(movementDirection3d, Matrix.CreateFromAxisAngle(Vector3.Normalize(axis), supportSlope));
                }



                float supportVelocityChange = Vector3.Dot(supportNormal, bodyShape.linearVelocity - Toolbox.getVelocityOfPoint(supportLocation, supportEntity));
                if (supportVelocityChange > maximumGlueSpeed)
                    isAirborne = true; //Getting pulled off the ground too much, let go!
                else
                {
                    //Remove the velocity along the support normal, don't want the character to fall through...
                    bodyShape.linearVelocity -= supportVelocityChange * supportNormal;

                    //Calculate all the information needed by the upcoming velocity changes.

                    Vector3 relativeVelocity = Toolbox.getVelocityOfPoint(supportLocation, bodyShape) - Toolbox.getVelocityOfPoint(supportLocation, supportEntity);
                    float relativeVelocityAlongMovement = Vector3.Dot(relativeVelocity, movementDirection3d);
                    float newVelocityAlongMovement = Math.Min(relativeVelocityAlongMovement + acceleration * dt, maximumSpeed);
                    float changeInVelocity = Math.Max(0, newVelocityAlongMovement - relativeVelocityAlongMovement);

                    //Push the character in the desired direction
                    bodyShape.linearVelocity += movementDirection3d * changeInVelocity;

                    //Decelerate any motion which is not in the direction of movement.

                    //Have to do some recalculations since the velocities have changed...
                    relativeVelocity = bodyShape.linearVelocity - Toolbox.getVelocityOfPoint(supportLocation, supportEntity);
                    Vector3 otherVelocity = relativeVelocity - MathHelper.Clamp(Vector3.Dot(relativeVelocity, movementDirection3d), 0, maximumSpeed) * movementDirection3d;
                    Vector3 otherVelocityDirection;
                    float otherVelocityMagnitude = otherVelocity.Length();
                    if (otherVelocityMagnitude > 0)
                        otherVelocityDirection = otherVelocity / otherVelocityMagnitude;
                    else
                        otherVelocityDirection = Vector3.Zero;
                    float otherVelocityRemoved = Math.Min(otherVelocityMagnitude, deceleration * dt);
                    bodyShape.linearVelocity -= otherVelocityDirection * otherVelocityRemoved;

                    //Apply forces on the support entity.
                    if (supportEntity.isPhysicallySimulated)
                    {
                        supportEntity.applyImpulse(supportLocation, supportEntityForceApplicationFactor * bodyShape.mass * supportVelocityChange * supportNormal         //Normal velocity change
                                                                  + supportEntityForceApplicationFactor * bodyShape.mass * otherVelocityRemoved * otherVelocityDirection //deceleration velocity change
                                                                  + supportEntityForceApplicationFactor * bodyShape.mass * -changeInVelocity * movementDirection3d);     //acceleration velocity change
                    }


                }
                

            }
            else
            {
                //Flying through the air.

                Vector3 movementDirection3d = new Vector3(movementDirection.X, 0, movementDirection.Y);
                float velocityMagnitude = bodyShape.linearVelocity.Length();
                Vector3 velocityDirection;
                if (velocityMagnitude > 0)
                    velocityDirection = bodyShape.linearVelocity / velocityMagnitude;
                else
                    velocityDirection = Vector3.Zero;
                float velocityAlongMovement = Vector3.Dot(bodyShape.linearVelocity, movementDirection3d);
                float newVelocityAlongMovement = Math.Min(velocityAlongMovement + airAcceleration * dt, maximumAirSpeed);
                float changeInVelocity = Math.Max(0, newVelocityAlongMovement - velocityAlongMovement);
                bodyShape.linearVelocity += movementDirection3d * changeInVelocity;
            }

            //Reposition the detectors for next frame.
            feetDetector.linearVelocity = bodyShape.linearVelocity;
            feetDetector.moveTo(bodyShape.centerPosition + feetDetectorOffset);
            bodyDetector.linearVelocity = bodyShape.linearVelocity;
            bodyDetector.moveTo(bodyShape.centerPosition);
        }

        

        /// <summary>
        /// Moves the character to the specified location.
        /// </summary>
        /// <param name="newPosition">Position to move the character to.</param>
        public void moveTo(Vector3 newPosition)
        {
            bodyShape.moveTo(newPosition);
            feetDetector.moveTo(newPosition + feetDetectorOffset);
            bodyDetector.moveTo(newPosition);
        }

        /// <summary>
        /// If supported, the character jumps into the air.
        /// </summary>
        public void jump()
        {
            if (!isAirborne)
            {
                isAirborne = true;
                bodyShape.linearVelocity += new Vector3(0, jumpSpeed, 0);
            }
        }


        /// <summary>
        /// Identifies a support, if any, for the character.
        /// </summary>
        /// <returns>Whether or not a support was found.</returns>
        bool findSupport()
        {

            float previousSupportAngle;
            if (!isAirborne)
                previousSupportAngle = (float)Math.Acos(Math.Min(1, Vector3.Dot(Vector3.Up, supportNormal)));
            else
                previousSupportAngle = 0;

            supportEntity = null;
            supportLocation = Toolbox.noVector;
            supportNormal = Toolbox.noVector;
            Vector3 sweepDown = new Vector3(0, -maximumStepHeight * 2, 0);
            float maximumToi;
            if (isAirborne)
            {

                //If it's flying through the air, don't try to 'step down' onto a support.
                maximumToi = .5f;
                //If it's flying upward, don't grab the ground with its feet.
                if (Vector3.Dot(bodyShape.linearVelocity, Vector3.Up) > 0)
                    return false;

            }
            else
                maximumToi = 1;
            
            Vector3 feetPosition = bodyShape.centerPosition + feetOffset;


            Vector3 locationCandidate;
            Vector3 normalCandidate;
            Vector3 newPosition = Vector3.Zero;

            float toi;
            float minimumToi = float.MaxValue;


            //First use the raytest to try to find something.
            foreach (Entity e in supportCandidates)
            {

                Vector3 ePosition = e.centerPosition;
                Quaternion eOrientation = e.orientationQuaternion;
         
                if (Toolbox.areSweptObjectsColliding(feetShape, e, ref feetPosition, ref ePosition, ref Toolbox.identityOrientation, ref eOrientation,  
                        0, e.collisionMargin, ref sweepDown, ref Toolbox.zeroVector, out locationCandidate, out normalCandidate, out toi) &&
                        toi < minimumToi && toi > 0 && toi <= maximumToi)
                {
                    normalCandidate.Normalize();
                    float dot = Vector3.Dot(normalCandidate, Vector3.Up);
                    if (dot < 0)
                    {
                        dot *= -1;
                        normalCandidate *= -1;
                    }
                    float supportAngle = (float)Math.Acos(Math.Min(1,dot));
                    newPosition = (toi - .5f) * sweepDown + bodyShape.centerPosition;
                    if(Math.Abs(toi - .5f) > .25f)
                    {//'Tall' step, verify source and target slopes
                        if (supportAngle > maximumStepSlope || previousSupportAngle > maximumStepSlope)
                           continue;
                    }
                    if (supportAngle < maximumSlope && (isNewPositionValid(newPosition)))
                    {
                        supportLocation = locationCandidate;
                        supportNormal = normalCandidate;
                        minimumToi = toi;
                        supportEntity = e;
                    }
                }
            }

            



            if (minimumToi < float.MaxValue)
            {
                isAirborne = false;
                moveTo(newPosition);
                return true;
            }
            else
            {
                isAirborne = true;
                return false;
            }

        }

        /// <summary>
        /// Determines if the new position is safe or will cause the character to bump its head or get stuck in anything.
        /// </summary>
        /// <param name="newPosition">Position to examine.</param>
        /// <returns>Whether or not the new position is valid.</returns>
        bool isNewPositionValid(Vector3 newPosition)
        {
            foreach (Entity e in nearbyEntities)
            {
                Vector3 ePosition = e.centerPosition;
                Quaternion eOrientation = e.orientationQuaternion;
                if (Toolbox.areObjectsCollidingMPR(bodyShape, e, ref newPosition, ref ePosition, ref Toolbox.identityOrientation, ref eOrientation, 0, e.collisionMargin))
                    return false;
            }
            return true;
        }

        /// <summary>
        /// Activates the character, adding its components to the space. 
        /// </summary>
        public void activate()
        {
            if (!isUpdating)
            {
                isUpdating = true;
                space.add(bodyDetector);
                space.add(bodyShape);
                space.add(feetDetector);
                isAirborne = true;
                bodyShape.linearVelocity = Vector3.Zero;
            }
        }

        /// <summary>
        /// Deactivates the character, removing its components from the space.
        /// </summary>
        public void deactivate()
        {
            if (isUpdating)
            {
                isUpdating = false;
                moveTo(new Vector3(10000, 0, 0));
                space.remove(bodyDetector);
                space.remove(bodyShape);
                space.remove(feetDetector);

            }
        }

        /// <summary>
        /// Called by the engine when the character is added to the space.
        /// Activates the character.
        /// </summary>
        /// <param name="newSpace">Space to which the character was added.</param>
        public override void addToSpace(Space newSpace)
        {
            base.addToSpace(newSpace); //sets this object's space to the newSpace.
            activate();
        }

        /// <summary>
        /// Called by the engine when the character is removed from the space.
        /// Deactivates the character.
        /// </summary>
        public override void removeFromSpace()
        {
            deactivate();
            base.removeFromSpace(); //Sets this object's space to null.
        }





    }
}
