/*
   BEPUphysicsDemos Project

      Copyright (C) 2008 Bepu Entertainment LLC

      This demonstration software, excluding the BEPUphysics.dll and BEPUphysics.xml (which are handled independently by their accompanying license),
      is provided 'as-is', without any express or implied warranty.  In no event will the authors be held liable for any damages
      arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
    Ross Nordby
    norbo@bepu-games.com
    Harrison Nordby
    squashwell@bepu-games.com
 */

using System;
using System.Collections.Generic;
using System.Text;
using BEPUphysics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;
using System.Diagnostics;

namespace BEPUphysicsDemos
{
    /// <remarks>
    /// Handles input and movement of a character in the game.
    /// Acts as a simple 'front end' for the bookkeeping and math of the character controller.
    /// </remarks>
    public class CharacterControllerInput
    {
        /// <summary>
        /// Physics representation of the character.
        /// </summary>
        public CharacterController characterController;
        /// <summary>
        /// Current offset from the position of the character to the 'eyes.'
        /// </summary>
        public Vector3 cameraOffset = new Vector3(0, .7f, 0);
        /// <summary>
        /// Camera to use for input.
        /// </summary>
        public Camera camera;
        /// <summary>
        /// Owning space of the character.
        /// </summary>
        public Space space;

        /// <summary>
        /// Whether or not to use the character controller's input.
        /// </summary>
        public bool isActive = true;


        /// <summary>
        /// Constructs the character and internal physics character controller.
        /// </summary>
        /// <param name="owningSpace">Space to add the character to.</param>
        /// <param name="cameraToUse">Camera to attach to the character.</param>
        public CharacterControllerInput(Space owningSpace, Camera cameraToUse)
        {
            characterController = new CharacterController(Vector3.Zero, 3, .5f, 5, .2f, 1.1f, 50, 70, 6, (float)Math.PI / 4, 5, 3, 2, 5, .1f, (float)Math.PI / 6, (float)Math.PI / 8);

            space = owningSpace;
            space.add(characterController);



            camera = cameraToUse;
            deactivate();
        }

        /// <summary>
        /// Gives the character control over the camera and movement input.
        /// </summary>
        public void activate()
        {
            if (!isActive)
            {
                isActive = true;
                camera.useMovementControls = false;
                characterController.activate();
                characterController.moveTo(camera.position);

            }

        }

        /// <summary>
        /// Returns input control to the camera.
        /// </summary>
        public void deactivate()
        {
            if (isActive)
            {
                isActive = false;
                camera.useMovementControls = true;
                characterController.deactivate();
            }
        }



        //Prevents spam-jumping; if the player jumps immediately after jumping, 
        //it is possible that they could still be in contact with the ground and thus get a double (or more) boost.
        float timeSinceLastJump = 0;
        /// <summary>
        /// Handles the input and movement of the character.
        /// </summary>
        /// <param name="dt">Time since last frame in simulation seconds.</param>
        public void update(float dt, KeyboardState keyboardInput, GamePadState gamePadInput)
        {
            if (isActive)
            {

                //Note that the character controller's update method is not called here; this is because it is handled within its owning space.
                //This method's job is simply to tell the character to move around based on the camera and input.

                //Puts the camera at eye level.
                camera.position = characterController.bodyShape.centerPosition + cameraOffset;
                Vector2 totalMovement = Vector2.Zero;
                timeSinceLastJump += dt;
#if XBOX360
                Vector3 forward = camera.getForwardVector();
                forward.Y = 0;
                forward.Normalize();
                Vector3 right = camera.getRightVector();
                totalMovement += gamePadInput.ThumbSticks.Left.Y * new Vector2(forward.X, forward.Z);
                totalMovement += gamePadInput.ThumbSticks.Left.X * new Vector2(right.X, right.Z);
                characterController.movementDirection = totalMovement;
#else


                //Collect the movement impulses.

                Vector3 movementDir;

                if (keyboardInput.IsKeyDown(Keys.E))
                {
                    movementDir = camera.getForwardVector();
                    totalMovement += Vector2.Normalize(new Vector2(movementDir.X, movementDir.Z));
                }
                if (keyboardInput.IsKeyDown(Keys.D))
                {
                    movementDir = camera.getForwardVector();
                    totalMovement -= Vector2.Normalize(new Vector2(movementDir.X, movementDir.Z));
                }
                if (keyboardInput.IsKeyDown(Keys.S))
                {
                    movementDir = camera.getLeftVector();
                    totalMovement += Vector2.Normalize(new Vector2(movementDir.X, movementDir.Z));
                }
                if (keyboardInput.IsKeyDown(Keys.F))
                {
                    movementDir = camera.getRightVector();
                    totalMovement += Vector2.Normalize(new Vector2(movementDir.X, movementDir.Z));
                }
                if (totalMovement == Vector2.Zero)
                    characterController.movementDirection = Vector2.Zero;
                else
                    characterController.movementDirection = Vector2.Normalize(totalMovement);

                //Jumping
                if (keyboardInput.IsKeyDown(Keys.A) && timeSinceLastJump > .3f)
                {
                    characterController.jump();
                    timeSinceLastJump = 0;
                }


#endif
            }
        }
    }
}
